/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: SendFile.cpp$
 * $Date: Sun Aug 10 06:36:50 1997$
 *
 Description:
\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <stdio.h>

#include "HandBase.h"
#include "HTTPCore.h"
#include "HTTPUtil.h"
#include "PIStrStr.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_HEADERPATTERN		"HeaderPattern"
#define KEY_CONF_FOOTERPATTERN		"FooterPattern"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	SendFile

Description:
	Send a file including headers to a remote client. If this is a 'HEAD'
	request only the HTTP headers will be sent.

<TR>
<TD>HeaderPattern
<TD>-
<TD>A Pi3Expression
<TD>Output first
<TD>HeaderPattern="<HTML><TITLE>The Title</TITLE>"

<TR>
<TD>FooterPattern
<TD>-
<TD>A Pi3Expression
<TD>Output last
<TD>FooterPattern="</HTML>"

<TR>
</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>

<H5>
	HeaderPattern
</H5>
Specifies a pattern of data that will be pre-appended to the file.

<H5>
	FooterPattern
</H5>
Specifies a pattern of data that will be appended to the file.

Phase:
	HANDLE

Returns:
	PIAPI_COMPLETED if the file was sent.
	If this handler was invoked for a phase other than HANDLE then PIAPI_ERROR
	is returned.

Note:
Example:
	<PRE>
	&ltObject&gt
		Name SendFile
		Class SendFileClass
	&lt/Object&gt

	&ltObject&gt
		...
		Handle SendFile
		...
	&lt/Object&gt

	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class SendFile : public HandlerBaseHTTP
{
private:
	/* ---
	Configuration data
	--- */
	Pi3Expression *pHeaderPattern;	/* pattern for top */
	Pi3Expression *pFooterPattern;	/* pattern for footer */

protected:
	/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *

				Load an expression or condition

	 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
	int LoadExpression( const char *pName, Pi3Expression **ppExpression,
		FnPi3Write *pFunctions, const char *pValue, PIOStrStream &os )	
		{
		assert( ppExpression );
		Pi3Expression *pExpression = *ppExpression;
		if ( pExpression )
			{
			os << "'" << pName << "' may only be specified once." << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};
		Pi3String *pError = Pi3String_new( 0 );
		pExpression = Pi3Expression_new( pValue, pFunctions, pError );
		if ( !pExpression )
			{
			os << "Error parsing expression: " <<
				Pi3String_getPtr( pError ) << ends;
			CONFIG_ERR( Object(), os.str() );
			Pi3Expression_delete( pExpression );
			Pi3String_delete( pError );
			return 0;
			};
		Pi3String_delete( pError );
		*ppExpression = pExpression;
		return 1;
		};

	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream os;
		os << pWhere << "SendFile: ";
		if ( !PIUtil_stricmp( KEY_CONF_HEADERPATTERN, pVariable ) )
			{
			if ( !LoadExpression( KEY_CONF_HEADERPATTERN, &pHeaderPattern,
				0, pValue, os ) )
				{ return 0; };
			assert( pHeaderPattern );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_FOOTERPATTERN, pVariable ) )
			{
			if ( !LoadExpression( KEY_CONF_FOOTERPATTERN, &pFooterPattern,
				0, pValue, os ) )
				{ return 0; };
			assert( pFooterPattern );
			}
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};

		return 1;
		};

public:
	SendFile( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	HandlerBaseHTTP( pObject ),
		pHeaderPattern( 0 ),
		pFooterPattern( 0 )
		{
		ReadParameters( iArgc, ppArgv );
		};

	~SendFile()
		{
		Pi3Expression_delete( pHeaderPattern );
		Pi3Expression_delete( pFooterPattern );
		};

	int Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer &/* tBuffer */ )
		{
		if ( iPhase!=PH_HANDLE )
			{ return PIAPI_ERROR; };

		PIDB *pR = tPIHTTP.pResponseDB;
		PIDB *pQ = tPIHTTP.pRequestDB;

		/* ---
		Get file object
		--- */
		const char *pPath = (const char *)PIDB_lookup( pR, PIDBTYPE_STRING,
			KEY_INT_PATH, 0 );
		PIFInfo *pFInfo = HTTPCore_getCachedFile( pPath ); 
		if ( !pFInfo )
			{ return PIAPI_ERROR; };

		/* ---	
		This file must exist and be a regular file 
		--- */
		if ( !PIFInfo_exists( pFInfo ) ||
			!PIFInfo_isRegular( pFInfo ) )
			{
			HTTPCore_logError( &tPIHTTP, "SendFile: File with \
path '%s' does not \
exist or is not a regular file.", PIFInfo_getPath( pFInfo ) );
			HTTPCore_releaseCachedFile( pFInfo );	
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
			};

		/* ---
		Lookup method 
		--- */
		int iMethod = (int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE,
			KEY_HTTP_METHOD, 0 );

		/* --- 	
		Get the header
		--- */
		enum { PATTERN_BUF_SIZE=1024 };
		char szHeader[PATTERN_BUF_SIZE];
		char *pHeader = 0;
		int iHeaderLen = 0;
		if ( pHeaderPattern )
			{
			pHeader = ( iMethod==MD_GET ) ? szHeader : 0;
			iHeaderLen = Pi3Expression_write( pHeaderPattern, &tPIHTTP, 0,
					pHeader, PATTERN_BUF_SIZE );
			if ( iHeaderLen>PATTERN_BUF_SIZE && pHeader )
				{
				pHeader = PI_NEW( char[iHeaderLen] );
				Pi3Expression_write( pHeaderPattern, &tPIHTTP, 0, pHeader,
					iHeaderLen );
				};
			};

		/* --- 	
		Get the footer
		--- */
		char szFooter[PATTERN_BUF_SIZE];
		char *pFooter = 0;
		int iFooterLen = 0;
		if ( pFooterPattern )
			{
			pFooter = ( iMethod==MD_GET ) ? szFooter : 0;
			iFooterLen = Pi3Expression_write( pFooterPattern, &tPIHTTP, 0,
					pFooter, PATTERN_BUF_SIZE );
			if ( iFooterLen>PATTERN_BUF_SIZE && pFooter )
				{
				pFooter = PI_NEW( char[iFooterLen] );
				Pi3Expression_write( pFooterPattern, &tPIHTTP, 0, pFooter,
					iFooterLen );
				};
			};

		/* --- content-length --- */
		enum { BUF_SIZE=63 };
		char szBuf[BUF_SIZE+1];
		unsigned long ulLen =  PIFInfo_getSize( pFInfo );
		ulLen += iHeaderLen;
		ulLen += iFooterLen;
		sprintf( szBuf, "%lu", ulLen );
		PIDB_add( pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH, szBuf, 0 );	

		/* --- last modified --- */
		Pi3String *pTmp = Pi3String_new( 0 );
		time_t tT = PIFInfo_getLastModified( pFInfo );
		PIPlatform_beforeUnsafeBlock();
		struct tm *pTms = gmtime( &tT );
		HTTPUtil_rFC822Time( pTms, pTmp );
		PIPlatform_afterUnsafeBlock();
		PIDB_add( pR, PIDBTYPE_RFC822, KEY_HTTP_LASTMODIFIED,
			(void *)Pi3String_getPtr( pTmp ), 0 );
		Pi3String_delete( pTmp );

		/* --- If current status is 0 then change to 200 OK --- */
		if ( !tPIHTTP.iStatus )
			{
			tPIHTTP.iStatus = ST_OK;
			};

		/* --- start response --- */
		if (	HTTPCore_sendGeneralHeaders( &tPIHTTP ) ||
				HTTPCore_sendEntityHeaders( &tPIHTTP, pR ) )
			{
			if ( pHeader && pHeader!=szHeader ) { PI_DELETE( [] pHeader ); };
			if ( pFooter && pFooter!=szFooter ) { PI_DELETE( [] pFooter ); };

			HTTPCore_releaseCachedFile( pFInfo );	
			return PIAPI_ERROR;
			};

		/* --- 
		If the Method is 'GET' return PIAPI_CONTINUE to let the 
		handling continue by sending the file.
		Otherwise assume Method is 'HEAD' and return PIAPI_COMPLETED as
		end of transaction
		--- */
		int iRet = 1;
		switch( iMethod )
			{
			case MD_GET:
				if ( pHeader && iHeaderLen )
					{
					iRet = ( PIIOBuffer_write( tPIHTTP.pBuffer, pHeader,
						iHeaderLen, PIIOBUF_NONE)>0 );
					};
				if ( iRet )
					{
					iRet = !HTTPUtil_sendFile( tPIHTTP.pBuffer, pFInfo,
						PIIOBUF_NONE, 0 );
					};
				if ( iRet && pFooter && iFooterLen )
					{
					iRet = ( PIIOBuffer_write( tPIHTTP.pBuffer, pFooter,
						iFooterLen, PIIOBUF_NONE)>0 );
					};
				break;
		
			case MD_HEAD:
				break;

			default:
				assert( 0 );
				HTTPCore_releaseCachedFile( pFInfo );	
				return PIAPI_ERROR;
			};

		if ( pHeader && pHeader!=szHeader ) { PI_DELETE( [] pHeader ); };
		if ( pFooter && pFooter!=szFooter ) { PI_DELETE( [] pFooter ); };

		/* ---
		If iRet is zero, an error occurred sending the file.
		** NOTE: Maybe do something with it.
		--- */

		HTTPCore_releaseCachedFile( pFInfo );	
		return PIAPI_COMPLETED;
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int SendFile_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseHTTP_constructor( pObj, PI_NEW( SendFile( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name SendFileClass
		Type LogicExtension
		Library HTTP
		OnClassLoad HandlerBaseHTTP_onClassLoad
		Constructor SendFile_constructor
		CopyConstructor HandlerBaseHTTP_copyConstructor
		Destructor HandlerBaseHTTP_destructor
		Execute HandlerBaseHTTP_execute
	</Class>

	<Object>
		Name SendFile
		Class SendFileClass
	</Object>

/*___+++CNF_END+++___*/
#endif

