/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: UNIXDmon.cpp$
 * $Date: Sun Aug 10 06:36:00 1997$
 *
 Description:
	UNIX daemon server class.
\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <assert.h>
#include <fstream.h>
#include <stdlib.h>
#include <string.h>

#if POSIX
#include <unistd.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#endif

#include "Pi2API.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_MESSAGE			"Message"
#define KEY_CONF_SERVEROBJECT		"ServerObject"
#define KEY_CONF_PIDFILE			"PIDFile"
#define KEY_CONF_RLIMITCORE			"RLimitCore"
#define KEY_CONF_RLIMITCPU			"RLimitCPU"
#define KEY_CONF_RLIMITDATA			"RLimitData"
#define KEY_CONF_RLIMITFSIZE		"RLimitFSize"
#define KEY_CONF_USER				"User"
#define KEY_CONF_GROUP				"Group"
#define KEY_CONF_NUMBEROFPROCESSES	"NumberOfProcesses"
#define KEY_CONF_LOCKFILENAME		"LockFileName"
#define KEY_CONF_MESSAGE			"Message"
#define CONFIG_ERR(this, msg)   \
	{ PILog_addMessage( PIObject_getDB(this), \
	PIObject_getConfigurationDB(this), PILOG_ERROR, (msg) ); }

/* ---
	Write error message to stderr, quoting this server as the 
	source
--- */
inline ostream &Cerr()
	{
	cerr << "UNIXDaemon[" << PIPlatform_getProcess() << ":" <<
		(int)PIThread_getSystemHandle( PIThread_getCurrent() ) << "]: ";
	return cerr;
	};

/*
#define D { Cerr() << __FILE__ << ", " << __LINE__ << endl; };
*/
#define D

/* ---
When a user and group is specified determine whether the main process
will change to the user, or only child processes
--- */
#define RUN_MAIN_AS_ROOT 0

/*____________________________________________________________________________*\
 *
 Description:
	Documentation.
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this server
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	UNIXDaemon

Description:
	This object performs some UNIX daemon actions, including changing the
	effective user and setting resource limits. This class also handles
	forking multiple child processes to handle requests.

Note:
	On non UNIX systems execution of this object will cause the object
	specified with 'ServerObject' to be executed after the message
	specified with 'Message' is displayed. All other options and actions
	are ignored.

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>ServerObject
<TD>+
<TD>a Pi3 object name
<TD>A Pi3 object
<TD>ServerObject "ServerObject"

<TR>
<TD>PIDFile
<TD>-
<TD>&ltfilepath&gt
<TD>Path to PID file
<TD>PIDFile "logs/pid.log"

<TR>
<TD>NumberOfProcesses
<TD>5
<TD>1..10000
<TD>Number of child processes
<TD>NumberOfProcesses 10

<TR>
<TD>RLimitCore
<TD>-
<TD>Number or 'Maximum'
<TD>Limit on corefile
<TD>RLimitCore 2000000

<TR>
<TD>RLimitCpu
<TD>-
<TD>Number or 'Maximum'
<TD>Limit on CPU time
<TD>RLimitCPU Maximum

<TR>
<TD>RLimitData
<TD>-
<TD>Number or 'Maximum'
<TD>Limit on data segment
<TD>RLimitData Maximum

<TR>
<TD>RLimitFSize
<TD>-
<TD>Number or 'Maximum'
<TD>Maximum filesize creatable
<TD>RLimitFSize Maximum

<TR>
<TD>User
<TD>-
<TD>username or #number
<TD>Child process user id
<TD>User jroy

<TR>
<TD>Group
<TD>-
<TD>groupname of #number
<TD>Child process group id
<TD>Group #-1

<TR>
<TD>LockFileName
<TD>-
<TD>&ltfilename&gt
<TD>server unique filename
<TD>LockFileName "Logs/lockfile.txt"

<TR>
<TD>Message
<TD>-
<TD>&lttext&gt
<TD>Startup message
<TD>Message "Welcome to OZ"

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>
	ServerObject
</H5>
Specify the server object which handles requests.

<H5>
	PIDFile
</H5>
Specifies a file to write the process identifier. This directive can be
repeated multiple times to create multiple PID files. This path can be
absolute or relative. If a relative path is given, the file will be 
created relative to the servers current directory.
<H5>
	NumberOfProcesses
</H5>
Gives the number of child processes to generate. As child processes
die, the parent monitor process will create new ones. When the
parent process dies, it will kill all child processes.
<H5>
RLimitCore
</H5>
<H5>
RLimitCPU
</H5>
<H5>
RLimitData
</H5>
<H5>
RLimitFSize
</H5>
Respectively reset the soft resource limit on corefile size, CPU time,
data segment size and largest creatable file for child processes of
this UNIX Daemon.
<H5>
User
</H5>
Specifies the user that child processes will execute as. A username can
be supplied or an id is assumed if the first letter is '#'
<H5>
Group
</H5>
Specifies the group that child processes will execute as. A groupname can
be supplied an id is assumed if the first letter is '#'.

<H5>
LockFileName
</H5>
Specifies a file into which this class will write a temporary filename
unique to this server and process group instance.
The path can be absolute or relative from the current directory. This
class will attempt to unlink the temporary file on exiting to this
directive should not be repeated more than once.

<H5>
Message
</H5>
Specifies a message to be written to standard output just before the
server enters its processing loop.

Returns:
	PIAPI_COMPLETED on success.
	PIAPI_CONTINUE if no action taken.
	PIAPI_ERROR and PIAPI_ABORT respectively for generic and severe
	error conditions.

Example:
	<PRE>
	&ltObject&gt
		Name UNIXDaemon
		Class UNIXDaemonClass
		ServerObject DefaultServerObject
		PIDFile	"Logs/pid.log"
		NumberOfProcesses 10
		RLimitCore Maximum
		RLimitCPU Maximum
		RLimitData Maximum
		RLimitFSize Maximum
		User jroy
		Group #-1
		LockFileName "Logs/lockfile.txt"
	&lt/Object&gt
	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

#define FLG_UID			0x0001
#define FLG_GID			0x0002
#define FLG_RCORE		0x0004
#define FLG_RCPU		0x0008
#define FLG_RDATA		0x0010
#define FLG_RFSIZE		0x0020

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class UNIXDaemon
{
private:
	PIObject *pObject;
	PIObject *pServerObject;
	int iOK;
	char *pMessage;

#if POSIX
	int iFlags; 
	uid_t tId;
	gid_t tGid;
	int iNumberOfProcesses;
	pid_t *pProcessIds;
	struct rlimit tRCore;
	struct rlimit tRCPU;
	struct rlimit tRData;
	struct rlimit tRFSize;
	char *pMainLockFileName;
	char *pLockFileName;
	pid_t tParentPid;			/* parents pid */

	/* --- lookup user id, returns 1 on success, else 0 --- */
	GetUid( const char *pValue )
		{
		assert( pValue );
		if ( !pValue )
			{ return 0; };

		/* --- must be root --- */
		if ( getuid() )
			{
			PILog_addMessage( PIObject_getDB( pObject ),
				PIObject_getConfigurationDB( pObject ), PILOG_ERROR,
				"%s", "UNIXDaemon: Must be root for 'User' directive." );
			return 0;
			};
	
		if ( *pValue == '#' )
			{ tId = atoi( &(pValue[1]) ); }
		else
			{ 
			int iFound = 0;
			struct passwd *pEnt = 0;
			PIPlatform_beforeUnsafeBlock();
			pEnt = getpwnam( pValue );
			if ( pEnt )
				{
				iFound = 1;
				tId = pEnt->pw_uid;
				};
			PIPlatform_afterUnsafeBlock();
			if ( !iFound ) 
				{
				PILog_addMessage( PIObject_getDB( pObject ),
					PIObject_getConfigurationDB( pObject ), PILOG_ERROR,
					"UNIXDaemon: Failed to get uid for user '%s'",
					pValue );
				return 0;
				};
			};
		
		iFlags |= FLG_UID;

		return 1;
		};

	/* --- lookup group id, returns 1 on success, else 0 --- */
	GetGid( const char *pValue )
		{
		assert( pValue );
		if ( !pValue )
			{ return 0; };
	
		/* --- must be root --- */
		if ( getuid() )
			{
			PILog_addMessage( PIObject_getDB( pObject ),
				PIObject_getConfigurationDB( pObject ), PILOG_ERROR,
				"%s", "UNIXDaemon: Must be root for 'Group' directive." );
			return 0;
			};
	
		if ( *pValue == '#' )
			{ tGid = atoi( &(pValue[1]) ); }
		else
			{ 
			int iFound = 0;
			struct group *pEnt = 0;
			PIPlatform_beforeUnsafeBlock();
			pEnt = getgrnam( pValue );
			if ( pEnt )
				{
				iFound = 1;
				tGid = pEnt->gr_gid;
				};
			PIPlatform_afterUnsafeBlock();
			if ( !iFound ) 
				{
				PILog_addMessage( PIObject_getDB( pObject ),
					PIObject_getConfigurationDB( pObject ), PILOG_ERROR,
					"UNIXDaemon: Failed to get gid for group '%s'", pValue );
				return 0;
				};
			};
		
		iFlags |= FLG_GID;

		return 1;
		};

protected:

	/* ---
	return 1 on success, 0 on failure
	--- */
	int AddResourceLimit( int iResource, const char *pName, const char *pValue,
		struct rlimit *pLimit )
		{
		assert( pName && pValue && pLimit );
		if ( !pName || !pValue || !pLimit )
			{ return 0; };

		/* --- get the limits --- */
		if ( ::getrlimit( iResource, pLimit )<0 )
			{
			PILog_addMessage( PIObject_getDB( pObject ),
				PIObject_getConfigurationDB( pObject ), PILOG_ERROR,
				"UNIXDaemon: getrlimit() failed for resource: '%s'", pName );
			return 0;
			};

		/* --- set the limits --- */
		if ( PIUtil_stricmp( pValue, "Maximum" ) )
			{ pLimit->rlim_cur = atoi( pValue ); }
		else
			{ pLimit->rlim_cur = pLimit->rlim_max; };

		return 1;
		};

	/* --- set the limits --- */
	int SetRLimit( int iResource, struct rlimit *pLimit, const char *pName )
		{
		assert( pLimit && pName );
		if ( !pLimit || !pName )
			{ return 0; };
		if ( ::setrlimit( iResource, pLimit )<0 )
			{
			PILog_addMessage( PIObject_getDB( pObject ),
				PIObject_getConfigurationDB( pObject ), PILOG_ERROR,
				"UNIXDaemon: setrlimit() failed for resource: '%s'",
				pName );
			Cerr() << "UNIXDaemon: setrlimit() failed for resource: '" <<
				pName << "'" << endl;
			return 0;
			};
		return 1;
		};

	/* ---
	Generate a temporary lockfile and write its name into the main
	lockfile.

	Return 1 on success, 0 on failure
	--- */
	int GenerateLockFile()
		{	
		if ( !*pLockFileName )
			{
			PILog_addMessage( PIObject_getDB( pObject ),
				PIObject_getConfigurationDB( pObject ), PILOG_ERROR,
				"UNIXDaemon: Could not generate lockfile name." );
			return 0;
			};
		ofstream ofs( pLockFileName );
		if ( ofs.good() )
			{
			ofs.close();

			/* --- file is readable and writable by this user only --- */
			::chmod( pLockFileName, S_IRUSR | S_IWUSR );
			}
		else
			{
			PILog_addMessage( PIObject_getDB( pObject ),
				PIObject_getConfigurationDB( pObject ), PILOG_ERROR,
				"UNIXDaemon: Could not create or change access to \
file '%s'.", pLockFileName );
			return 0;
			};
		return 1;
		}
#endif

	/* ---
	Parameters 
	--- */
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
        assert( pVariable && pValue );
        if ( !pVariable || !pValue )
            { return 0; };

        if ( !PIUtil_stricmp( KEY_CONF_SERVEROBJECT, pVariable ) )
            {
			/* ---
			loaded in constructor after all parameters have been read
			--- */
            }
        else if ( !PIUtil_stricmp( KEY_CONF_MESSAGE, pVariable ) )
			{
			pMessage = PIUtil_strdup( pValue );
			}
#if POSIX
        else if ( !PIUtil_stricmp( KEY_CONF_NUMBEROFPROCESSES, pVariable ) )
			{
			int iTmp = atoi( pValue );
			if ( iTmp<1 || iTmp>10000 )	/* if this changed, then change the	
										documentation above */
				{
				PILog_addMessage( PIObject_getDB( pObject ),
					PIObject_getConfigurationDB( pObject ), PILOG_ERROR,
					"%sUNIXDaemon: Number of processes must be between 1 \
and 10,000, the value specified is: '%d'", pWhere, iTmp );
				return 0;
				};

			iNumberOfProcesses = iTmp;
			}
        else if ( !PIUtil_stricmp( KEY_CONF_RLIMITCORE, pVariable ) )
			{
			if ( AddResourceLimit( RLIMIT_CORE, "RLIMIT_CORE", pValue,
				&tRCore ) )
				{ iFlags |= FLG_RCORE; }
			else
				{ return 0; };
			}
        else if ( !PIUtil_stricmp( KEY_CONF_RLIMITCPU, pVariable ) )
			{
			if ( AddResourceLimit( RLIMIT_CORE, "RLIMIT_CORE", pValue,
				&tRCPU ) )
				{ iFlags |= FLG_RCPU; }
			else
				{ return 0; };
			}
        else if ( !PIUtil_stricmp( KEY_CONF_RLIMITDATA, pVariable ) )
			{
			if ( AddResourceLimit( RLIMIT_CORE, "RLIMIT_DATA", pValue,
				&tRData ) )
				{ iFlags |= FLG_RDATA; }
			else
				{ return 0; };
			}
        else if ( !PIUtil_stricmp( KEY_CONF_RLIMITFSIZE, pVariable ) )
			{
			if ( AddResourceLimit( RLIMIT_CORE, "RLIMIT_FSIZE", pValue,
				&tRFSize ) )
				{ iFlags |= FLG_RFSIZE; }
			else
				{ return 0; };
			}
        else if ( !PIUtil_stricmp( KEY_CONF_USER, pVariable ) )
			{
			if ( !GetUid( pValue ) )
				{ return 0; };
			}
        else if ( !PIUtil_stricmp( KEY_CONF_GROUP, pVariable ) )
			{
			if ( !GetGid( pValue ) )
				{ return 0; };
			}
        else if ( !PIUtil_stricmp( KEY_CONF_PIDFILE, pVariable ) )
            {
			ofstream ofs( pValue );
			if ( ofs.good() )
				{
				ofs << ::getpid() << endl;
				}
			else
				{
				PILog_addMessage( PIObject_getDB( pObject ),
					PIObject_getConfigurationDB( pObject ), PILOG_ERROR,
					"%sUNIXDaemon: Unable to open PID file: '%s'",
					pWhere, pValue );
				return 0;
				};
			}
        else if ( !PIUtil_stricmp( KEY_CONF_LOCKFILENAME, pVariable ) )
			{
			const char *pTmp = tempnam( 0, "Pi3" );
			assert( pTmp );
			pLockFileName = PIUtil_strdup( pTmp );

			/* ---
			--- */
			assert( pValue );
			assert( pLockFileName );
			ofstream ofs( pValue );
			if ( ofs.good() ) {
				ofs << pLockFileName << endl;
				pMainLockFileName = PIUtil_strdup( pValue );
				}
			else
				{
				PILog_addMessage( PIObject_getDB( pObject ),
					PIObject_getConfigurationDB( pObject ), PILOG_ERROR,
					"UNIXDaemon: Could not write lockfile name '%s' \
into specified file '%s'.", pLockFileName, pValue );
				*pLockFileName = '\0';
				return 0;
				};
			}
		else
			{
			PILog_addMessage( PIObject_getDB( pObject ),
				PIObject_getConfigurationDB( pObject ), PILOG_ERROR,
				"%sUNIXDaemon: Unknown parameter '%s'", pWhere, pVariable );
			};
#endif

		return 1;
		};

	/* --- --- */
    static int ParameterFn( void *pData, const char *pVar, const char *pVal,
		const char *pWhere )
        {
        assert( pData );
        return ((UNIXDaemon *)pData)->Parameter( pVar, pVal, pWhere );
        };

public:
	/* --- --- */
	UNIXDaemon( PIObject *pTheObject, int iArgc, const char *ppArgv[] )
	:	pObject( pTheObject ),
		pServerObject( 0 ),
		iOK( 0 ),
		pMessage( 0 )
#if POSIX
		, iFlags( 0 ),
		tId( (uid_t)-1 ),
		tGid( (gid_t)-1 ),
		iNumberOfProcesses( 5 ),	/* if this is changed then change the
									documentation above */
		pProcessIds( 0 ),
		pMainLockFileName( 0 ),
		pLockFileName( 0 ),
		tParentPid( ::getpid() )
		{
		if ( !PIObject_readParameters( pObject, iArgc, ppArgv, ParameterFn,
			this ) )
			{ return; };

		/* --- Generate lockfile ? --- */
		if ( pLockFileName )
			{
			if ( !GenerateLockFile() )
				{ return; };
			};

		pProcessIds = new pid_t[iNumberOfProcesses];
		if ( !pProcessIds )
			{ return; };

		memset( pProcessIds, 0, sizeof( pid_t ) * iNumberOfProcesses );
#else
		{
		if ( !PIObject_readParameters( pObject, iArgc, ppArgv, ParameterFn,
			this ) )
			{ return; };
#endif

		/* --- load up the server object --- */
		pServerObject = PIObject_loadEx( pObject, KEY_CONF_SERVEROBJECT );
		if ( !pServerObject )
			{ return; };
		iOK = 1;
		};

	/* --- --- */
	virtual ~UNIXDaemon()
		{
#if POSIX
		/* ---
		These cleanup tasks shutdown the server cleanly, this is not
		something that should happen when a child process exists, so
		return if this is a child process.
		We can determine if this is a child process by comparing the
		the current process id to tParentPid, the process id when
		this object was constructed and before any fork's.	
		--- */
		if ( ::getpid()!=tParentPid )
			{ return; };

		int i;

		/* --- kill child processes --- */
		if ( pProcessIds )
			{
			for( i=0; i<iNumberOfProcesses; i++ )	
				{
				if ( pProcessIds[i] )
					{
					::kill( pProcessIds[i], SIGTERM );
					};
				};
			};

		/* ---
		NOTE: maybe wait for notification that processes have died
		--- */

		/* --- remove the file containing the lockfile name --- */
		if ( pMainLockFileName && *pMainLockFileName )
			{ ::unlink( pMainLockFileName ); };
		if ( pMainLockFileName )
			{ PIUtil_free( pMainLockFileName ); };

		/* --- remove the lockfile --- */
		if ( pLockFileName && *pLockFileName )
			{ ::unlink( pLockFileName ); };
		if ( pLockFileName )
			{ PIUtil_free( pLockFileName ); };
		delete [] pProcessIds;
#endif
		PIObject_delete( pServerObject, 0, 0 );
		if ( pMessage ) { PIUtil_free( pMessage ); };
		};

#if POSIX
	/* --- signal handler --- */
	static void SignalHandler( int iSignal )
		{
		switch( iSignal )
			{
			case SIGTERM:
				exit( 0 );
				break;

			default:;
			};
		};

	/* --- child function --- */
	int ChildMain( int iArgc, const char *ppArgv[] )
		{
		/* --- ignore SIGINT --- */
		signal( SIGINT, SIG_IGN );

		/* --- quick exit on sigterm, don't bother freeing memory --- */
		signal( SIGTERM, SignalHandler );

		/* --- make sure we don't kill any siblings --- */
		memset( pProcessIds, 0, sizeof( pid_t ) * iNumberOfProcesses );

		/* --- set resource limits --- */
		if ( iFlags & FLG_RCORE )
			{
			if ( !SetRLimit( RLIMIT_CORE, &tRCore, "RLIMIT_CORE" ) )
				{ return PIAPI_ABORT; };
			};
		if ( iFlags & FLG_RCPU )
			{
			if ( !SetRLimit( RLIMIT_CPU, &tRCPU, "RLIMIT_CPU" ) )
				{ return PIAPI_ABORT; };
			};
		if ( iFlags & FLG_RDATA )
			{
			if ( !SetRLimit( RLIMIT_DATA, &tRData, "RLIMIT_DATA" ) )
				{ return PIAPI_ABORT; };
			};
		if ( iFlags & FLG_RFSIZE )
			{
			if ( !SetRLimit( RLIMIT_FSIZE, &tRFSize, "RLIMIT_FSIZE" ) )
				{ return PIAPI_ABORT; };
			};

#if RUN_MAIN_AS_ROOT
		/*
		** Change the effective user and group ids
		*/
		/* --- group id --- */
		if ( iFlags & FLG_GID )
			{
			if ( ::setgid( tGid )==-1 )
				{
				Cerr() << "Could not set group id to '" << tGid <<
					"' error code is " << errno << endl;
				return PIAPI_ABORT;
				};
			};

		/* --- user id --- */
		if ( iFlags & FLG_UID )
			{
			if ( ::setuid( tId )==-1 )
				{
				Cerr() << "Could not set user id to '" << tId <<
					"' error code is " << errno << endl;
				return PIAPI_ABORT;
				};
			};
#endif

		int iRet = PILogic_execute( pServerObject, iArgc, ppArgv );

		/* ---
		If there was an error write out details
		--- */
		if ( iRet )
			{
	        Cerr() << "Error(s) during child execution:-" << endl;
			PILog_writeDiagnostics( PIObject_getDB( pObject ), PILOG_ALL, 0 );
			cout.flush();
			cerr.flush();
			};

		return 0;
		}; 
#endif

	/* --- --- */
	int DoExecute( int iArgc, const char *ppArgv[] )
		{
		/* --- Write start message? --- */
		if ( pMessage )
			{
			cout << pMessage << endl;
			};
#if	!POSIX
		return PILogic_execute( pServerObject, iArgc, ppArgv );
#else
		assert( pProcessIds );

#if !RUN_MAIN_AS_ROOT
		/*
		** Change the effective user and group ids
		*/
		/* --- group id --- */
		if ( iFlags & FLG_GID )
			{
			if ( ::setgid( tGid )==-1 )
				{
				Cerr() << "Could not set group id to '" << tGid <<
					"' error code is " << errno << endl;
				return PIAPI_ERROR;
				};
			};

		/* --- user id --- */
		if ( iFlags & FLG_UID )
			{
			if ( ::setuid( tId )==-1 )
				{
				Cerr() << "Could not set user id to '" << tId <<
					"' error code is " << errno << endl;
				return PIAPI_ABORT;
				};
			};
#endif

		int iCurrentNumberOfChildren = 0;

		for(;;)
			{
			/* --- create new child processes --- */
			while( iCurrentNumberOfChildren<iNumberOfProcesses )
				{
#if 0
/*
** LATER, XXX
**
** This is obselete
*/
				/* ---
				If a lockfile was created then before forking
				check that the lockfile still exists, a child
				may have removed it in an attempt to stop the server 
				because of a tight fork() loop caused by a problem with
				the lockfile
				--- */
				struct stat tStat;
				if ( pLockFileName && stat( pLockFileName, &tStat )==-1 )
					{
					Cerr() << "**" << endl;
					Cerr() << "** Lockfile '" << pLockFileName <<
						"' does not exist, aborting daemon." << endl;
					Cerr() << "**" << endl;
					Cerr() << "** This typically indicates a serious error in \
child process execution." << endl;
					Cerr() << "**" << endl;
					return PIAPI_ABORT;
					};
#endif

				int iChildPid = ::fork();
				switch( iChildPid )
					{
					case -1:
						/* --- assume this is EAGAIN for tables full --- */
						sleep( 5 );		/* extremely sophisticated response */
						break;

					case 0:
						if ( ChildMain( iArgc, ppArgv )==PIAPI_ABORT )
							{
							/* --- Abort the server --- */
							if ( pLockFileName )
								{ ::unlink( pLockFileName ); };
							};
						return PIAPI_COMPLETED;

					default:
						/* ---
						this is the parent, record this new child.
						--- */
						{
						iCurrentNumberOfChildren++;
						assert( iCurrentNumberOfChildren <=
							iNumberOfProcesses );
						int i=0;
						for(i=0; i<iNumberOfProcesses; i++ )
							{
							if ( pProcessIds[i]==0 )
								{
								pProcessIds[i] = iChildPid;
								break;
								};
							};
						assert( i<iNumberOfProcesses );
						};
					};	/* --- switch on fork --- */
				}; /* --- loop while we have not reached max.# processes --- */ 

			/* ---
			Before we can create another process we must wait for one
			to die
			--- */
			for(;;)
				{
				int iStatus;
				pid_t tPid = ::wait( &iStatus );
				if ( tPid==-1 )
					{
					if ( errno!=EINTR )
						{
						/* --- signal caught, exit --- */
						Cerr() << "Unknown error from wait(): " << errno
							<< endl;
						};
					return PIAPI_COMPLETED;
					}
				else
					{
					/* ---
					if this child process did not exit normally, then
					log a message 
					--- */
					if ( !(WIFEXITED(iStatus)) )
						{
						Cerr() << "Child process " << tPid <<
							" did not exit normally, status: " << iStatus
							<< endl;
						};

					/* --- mark this process as reaped --- */
					int iProcessSlot = -1;	/* slot of exitted process */
					for(int i=0; i<iNumberOfProcesses; i++ )
						{
						if ( pProcessIds[i]==tPid )
							{
							iProcessSlot = i;
							pProcessIds[i] = 0;
							break;
							};
						};
					if ( iProcessSlot==-1 )
						{
						/* --- 
						This is not a child process we know about. This
						could happen if a restart occurred (via SIGHUP) and
						child processes from a previous iteration are 
						signalling.

						In this case ignore the signal
						---- */
						}	
					else
						{
						/* ----
						One of the current children has died, decrement the
						child process count and break to allow a new one
						to be created
						--- */
						iCurrentNumberOfChildren--;
						break;
						};
					};
				};
			};

		return PIAPI_COMPLETED;
#endif	/* POSIX */
		};
	
	/* --- --- */
	int Execute( int iArgc, const char *ppArgv[] )
		{
		int iRet = DoExecute( iArgc, ppArgv );
		switch( iRet )
			{
			case PIAPI_COMPLETED:
				break;


			case PIAPI_ABORT:
				CONFIG_ERR( pObject, "UNIXDaemon: A critical error occurred." );
				break;

			default:
				CONFIG_ERR( pObject, "UNIXDaemon: An error occurred." );
			};
		return iRet;
		};

	/* --- --- */
	inline int IsOK()		{ return iOK; };
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int UNIXDaemon_onClassLoad( PIClass_LoadAction eLoad, int,
	const char *[] )
{
	switch( eLoad )
		{
		case STARTUP:
		default:;
		};

	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int UNIXDaemon_constructor( PIObject *pObj, int iArgc,
		const char *ppArgv[] )
{
	UNIXDaemon *pUNIXDaemon = new UNIXDaemon( pObj, iArgc, ppArgv );
	if (!( pUNIXDaemon && pUNIXDaemon->IsOK() ))
		{
		delete pUNIXDaemon;
		return PIAPI_ERROR;
		};
	PIObject_setUserData( pObj, pUNIXDaemon );
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int UNIXDaemon_copyConstructor( PIObject *, int, const char *[] )
{
	/* --- this class doesn't support copy construction --- */
	return PIAPI_ERROR;	
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int UNIXDaemon_execute( PIObject *pObj, int iArgc, 
	const char *ppArgv[] )
{
	if ( !pObj ) return PIAPI_ERROR;
	return ((UNIXDaemon *)PIObject_getUserData(pObj))->Execute( iArgc, ppArgv );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int UNIXDaemon_destructor( PIObject *pObj, int, const char *[] )
{
	delete (UNIXDaemon *)PIObject_getUserData( pObj );
	return PIAPI_COMPLETED;
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name UNIXDaemonClass
		Type LogicExtension
		Library Server
		OnClassLoad UNIXDaemon_onClassLoad
		Constructor UNIXDaemon_constructor
		CopyConstructor UNIXDaemon_copyConstructor
		Destructor UNIXDaemon_destructor
		Execute UNIXDaemon_execute
	</Class>

	<Object>
		Name UNIXDaemon
		Class UNIXDaemonClass
	</Object>

/*___+++CNF_END+++___*/
#endif

